<?php
/**
 * Perforce Swarm
 *
 * @copyright   2013-2018 Perforce Software. All rights reserved.
 * @license     Please see LICENSE.txt in top-level readme folder of this distribution.
 * @version     2018.1/1742263
 */

namespace Api\Controller;

use Api\AbstractApiController;
use P4\Exception;
use Zend\Http\Response;
use Zend\View\Model\JsonModel;

/**
 * Swarm Login API
 *
 * @SWG\Resource(
 *   apiVersion="v9",
 *   basePath="/api/v9/"
 * )
 */
class LoginController extends AbstractApiController
{
    const RESULTS = 'results';
    const METHOD  = 'method';
    const METHODS = 'methods';
    const CODE    = 'code';
    const OPTION  = 'option';
    const ERROR   = 'error';
    const TOKEN   = 'token';
    const POLL    = 'poll';
    const PROMPT  = 'prompt';
    const TRIGGER = 'trigger';

    const NEXTSTATE  = 'nextState';
    const SUCCESSMSG = 'successMsg';
    const MFASTATUS  = 'mfaStatus';
    const CHALLENGE  = 'challenge';

    const RESPONSEMETHOD = 'responseMethod';

    const MFASUCCESS     = "mfaSuccess";
    const MFAPOLLPENDING = "mfaPending";
    const MFAFAILED      = "mfaFailed";
    const CHECK_AUTH     = 'check-auth';

     /**
     * @SWG\Api(
     *     path="listmethods/",
     *     @SWG\Operation(
     *         method="GET",
     *         summary="Get List of 2FA Methods",
     *         notes="Returns the complete list of methods of 2FA.",
     *         nickname="listMethods"
     *     )
     * )
     *
     * @apiUsageExample Listing 2FA Methods
     *
     *   To list the 2FA methods:
     *
     *   ```bash
     *   curl -u "username:password" \
     *        "https://myswarm.url/api/v9/login/listmethods"
     *   ```
     *
     *   Swarm responds with a list of 2FA methods:
     *
     *   ```json
     *   {
     *     "results": {
     *       "methods": {
     *         "1": {
     *           "methodName": "Method Name will be here",
     *           "methodDesc": "Method Description will be here"
     *         },
     *         "2": {
     *           "methodName": "Method Name will be here",
     *           "methodDesc": "Method Description will be here"
     *         },
     *         "3": {
     *           "methodName": "Method Name will be here",
     *           "methodDesc": "Method Description will be here"
     *         },
     *         "4": {
     *           "methodName": "Method Name will be here",
     *           "methodDesc": "Method Description will be here"
     *         }
     *       }
     *     },
     *     "option": {
     *       "persist": "option",
     *       "nextState": "init-auth"
     *     },
     *     "code": 200
     *   }
     *   ```
     *
     * @apiSuccessExample Successful Response:
     *     HTTP/1.1 200 OK
     *
     *     {
     *       "results": {
     *         "methods": {
     *           "1": {
     *             "methodName": "Method Name will be here",
     *             "methodDesc": "Method Description will be here"
     *           },
     *           "2": {
     *             "methodName": "Method Name will be here",
     *             "methodDesc": "Method Description will be here"
     *           },
     *           "3": {
     *             "methodName": "Method Name will be here",
     *             "methodDesc": "Method Description will be here"
     *           },
     *           "4": {
     *             "methodName": "Method Name will be here",
     *             "methodDesc": "Method Description will be here"
     *           }
     *         }
     *       },
     *       "option": {
     *         "persist": "option",
     *         "nextState": "init-auth"
     *       },
     *       "code": 200
     *     }
     *
     * @return mixed
     */
    public function listMethodsAction()
    {
        $services = $this->getServiceLocator();
        $p4User   = $services->get('p4_user');

        // Set the flags to get list of methods
        $flags = array('-S', 'list-methods');

        try {
            // Run the command to fetch the data
            $commandResults = $p4User->run('login2', $flags, null, true);
            $commandData    = $commandResults->getData();
            $option         = $commandData[1];
            // If we don't get returned data[1] then we can assume the command failed.
            if (empty($option)) {
                if (strpos($commandData[0], 'does not use second factor authentication.') !== false) {
                    return new JsonModel(
                        array(
                            static::CODE    => 200,
                            static::RESULTS => $commandData[0]
                        )
                    );
                } else {
                    throw new Exception("Missing required data.");
                }
            }
            // Collect methods and option
            $options = $option;
            $methods = $commandData[0];

            // If we are returned 1 or more methods return them to the page.
            if (count($methods) > 0) {
                $results[static::RESULTS][static::METHODS] = $this->listMethodSort($methods);
            }

            // Set the option and code vaules to be returned.
            $results[static::OPTION] = $options;
            $results[static::CODE]   = Response::STATUS_CODE_200;

            return $this->prepareSuccessModel($results);
        } catch (Exception $e) {
            // set message to the error from P4 and just return it.
            $message = $e->getMessage();
        }

        // If we hit here then we have failed and should return the error if we have one.
        return new JsonModel(
            array(
                'isValid'     => false,
                static::CODE  => Response::STATUS_CODE_401,
                static::ERROR => $message
            )
        );
    }

    /**
     * @SWG\Api(
     *     path="initauth/",
     *     @SWG\Operation(
     *         method="POST",
     *         summary="initiating the 2FA authentication",
     *         notes="initiating the 2FA authentication",
     *         nickname="initAuth",
     *         @SWG\Parameter(
     *             name="method",
     *             description="The Method in which you want to use.",
     *             paramType="form",
     *             type="string",
     *             required=true
     *         )
     *     )
     * )
     *
     * @apiUsageExample Initiating the 2FA authentication
     *
     *   To Initiate the user 2FA login:
     *
     *   ```bash
     *   curl -X POST -u "username:password" \
     *        -d "method=METHOD" \
     *        "https://myswarm.url/api/v9/login/initauth"
     *   ```
     *
     *   Assuming that the authenticated user has permission, Swarm responds with the next step in the 2FA login:
     *
     *   ```json
     *   {
     *     "results": {
     *       "trigger": "TriggerName",
     *       "successMsg": "Message from Authentication method"
     *     },
     *     "option": {
     *       "prompt": true,
     *       "nextState": "check-auth"
     *     },
     *     "code": 200
     *   }
     *   ```
     *
     * @apiSuccessExample Successful Response:
     *     HTTP/1.1 200 OK
     *
     *     {
     *       "results": {
     *          "trigger": "TriggerName",
     *          "successMsg": "Message from Authentication method"
     *       },
     *       "option": {
     *         "prompt": true,
     *         "nextState": "check-auth"
     *       },
     *       "code": 200
     *     }
     *
     * @param POST
     * @return JsonModel
     */
    public function initAuthAction()
    {
        $request   = $this->getRequest();
        $services  = $this->getServiceLocator();
        $p4User    = $services->get('p4_user');
        $method2fa = $request->getPost(static::METHOD);

        // Set the flags to init auth method
        $flags = array('-S', 'init-auth', '-m', $method2fa);

        try {
            // Run the command to fetch the data
            $commandResults = $p4User->run('login2', $flags, null, true);
            $initAuthData   = $commandResults->getData(0);
            if (!is_array($initAuthData)
                && strpos($initAuthData, 'does not use second factor authentication.') !== false) {
                return new JsonModel(
                    array(
                        static::CODE    => 200,
                        static::RESULTS => $initAuthData
                    )
                );
            }

            // Set the trigger and successmsg that we need to display.
            $results[static::RESULTS] = array(
                static::TRIGGER    => $initAuthData[static::TRIGGER],
                static::SUCCESSMSG => $initAuthData[static::SUCCESSMSG]
            );

            if (isset($initAuthData[static::CHALLENGE])) {
                $results[static::RESULTS][static::CHALLENGE] = $initAuthData[static::CHALLENGE];
            }

            // The option that tells us if we are prompt or poll
            $results[static::OPTION] = array(
                static::RESPONSEMETHOD => $initAuthData[static::RESPONSEMETHOD],
                static::NEXTSTATE      => $initAuthData[static::NEXTSTATE]
            );

            // Set the code to successful.
            $results[static::CODE] = Response::STATUS_CODE_200;

            return $this->prepareSuccessModel($results);
        } catch (Exception $e) {
            // set message to the error from P4 and just return it.
            $message = $e->getMessage();
        }
        // If we hit here then we have failed and should return the error if we have one.
        return new JsonModel(
            array(
                'isValid'     => false,
                static::CODE  => Response::STATUS_CODE_401,
                static::ERROR => $message
            )
        );
    }

    /**
     * @SWG\Api(
     *     path="checkauth/",
     *     @SWG\Operation(
     *         method="POST",
     *         summary="checking the 2FA authentication",
     *         notes="checking the 2FA authentication",
     *         nickname="checkAuth",
     *         @SWG\Parameter(
     *             name="token",
     *             description="The token from the user for their 2FA prompt.",
     *             paramType="form",
     *             type="string",
     *             required=true
     *         )
     *     )
     * )
     *
     * @apiUsageExample Checking the 2FA authentication
     *
     *   To Check User prompt input for 2FA
     *
     *   ```bash
     *   curl -X POST -u "username:password" \
     *        -d "token=TOKEN" \
     *        "https://myswarm.url/api/v9/login/checkauth"
     *   ```
     *
     *   Assuming that the authenticated user has permission, Swarm responds with the next step in the 2FA login:
     *
     *   ```json
     *   {
     *     "results": {
     *       "trigger": "GAuth says yes!",
     *       "successMsg": "Second factor authentication approved."
     *     },
     *     "code": 200
     *   }
     *   ```
     *
     * @apiSuccessExample Successful Response:
     *     HTTP/1.1 200 OK
     *
     *     {
     *       "results": {
     *         "trigger": "otp-generated|||GAuth says yes!",
     *         "successMsg": "Second factor authentication approved."
     *       },
     *       "code": 200
     *     }
     *
     * @param POST
     * @return JsonModel
     */
    public function checkAuthAction()
    {
        $request  = $this->getRequest();
        $services = $this->getServiceLocator();
        $p4User   = $services->get('p4_user');
        $prompt   = $request->getPOST(static::TOKEN);

        // Set the flags to check auth
        $flags = array('-S', static::CHECK_AUTH);

        if ($prompt !== null) {
            try {
                // Run the command to fetch the data
                $commandResults = $p4User->run('login2', $flags, $prompt, true);
                $checkAuthData  = $commandResults->getData();

                // Ensure the data has been set.
                if ($checkAuthData === false) {
                    $checkAuthData = $commandResults->getData(0);
                }
                // Set the code and results to be returned.
                $results[static::RESULTS]   = $checkAuthData;
                $results[static::CODE]      = Response::STATUS_CODE_200;
                $results[static::MFASTATUS] = static::MFASUCCESS;

                return $this->prepareSuccessModel($results);
            } catch (Exception $e) {
                // set message to the error from P4 and just return it.
                $message = $e->getMessage();
            }
        } else {
            try {
                $commandResults = $p4User->run('login2', $flags, $prompt, true);
                $checkAuthData  = $commandResults->getData();
                $message        = $checkAuthData[0];
                if (strpos($message, 'does not use second factor authentication.') !== false) {
                    return new JsonModel(
                        array(
                            static::CODE      => 200,
                            static::RESULTS   => $message,
                            static::MFASTATUS => static::MFASUCCESS
                        )
                    );
                }
            } catch (Exception $e) {
                // set message to the error from P4 and just return it.
                $message = $e->getMessage();
            }
        }
        // If we hit here then we have failed and should return the error if we have one.
        return new JsonModel(
            array(
                'isValid'         => false,
                static::CODE      => Response::STATUS_CODE_401,
                static::ERROR     => $message,
                static::MFASTATUS => static::MFAFAILED
            )
        );
    }

    /**
     * @SWG\Api(
     *     path="checkauth/",
     *     @SWG\Operation(
     *         method="GET",
     *         summary="checking the 2FA authentication",
     *         notes="checking the 2FA authentication",
     *         nickname="checkAuthPoll"
     *     )
     * )
     *
     * @apiUsageExample Checking the 2FA authentication
     *
     *   To Check User prompt input for 2FA
     *
     *   ```bash
     *   curl -u "username:password" \
     *        "https://myswarm.url/api/v9/login/checkauth"
     *   ```
     *
     *   Assuming that the authenticated user has permission, Swarm responds with the next step in the 2FA login:
     *
     *   ```json
     *   {
     *     "results": {
     *       "trigger": "GAuth says yes!",
     *       "successMsg": "Second factor authentication approved."
     *     },
     *     "code": 200
     *   }
     *   ```
     *
     * @apiSuccessExample Successful Response:
     *     HTTP/1.1 200 OK
     *
     *     {
     *       "results": {
     *         "trigger": "GAuth says yes!",
     *         "successMsg": "Second factor authentication approved."
     *       },
     *       "code": 200
     *     }
     *
     * @return JsonModel
     */
    public function checkAuthPollAction()
    {
        $services = $this->getServiceLocator();
        $p4User   = $services->get('p4_user');
        $message  = '';

        // Set the flags to check auth poll
        $flags = array('-S', static::CHECK_AUTH);
        try {
            // Run the command to fetch the data
            $commandResults    = $p4User->run('login2', $flags, null, true);
            $checkAuthPollData = $commandResults->getData();

            // Set the code and results to be returned.
            $results[static::RESULTS]   = $checkAuthPollData;
            $results[static::CODE]      = Response::STATUS_CODE_200;
            $results[static::MFASTATUS] = static::MFASTATUS;

            return $this->prepareSuccessModel($results);
        } catch (Exception $e) {
            // set message to the error from P4 and just return it.
            $message  = $e->getMessage();
            $pending  = (strpos($e->getMessage(), "Second factor authentication approval is pending.") > 0);
            $rejected = (strpos($e->getMessage(), "Second factor authentication was rejected.") > 0);

            if ($pending || $rejected) {
                $checkAuthPollData = array_merge($e->getResult()->getData(), $e->getResult()->getErrors());

                // Set the code and results to be returned.
                $results[static::RESULTS] = $checkAuthPollData;
                $results[static::CODE]    = Response::STATUS_CODE_200;

                if ($pending) {
                    $results[ static::MFASTATUS ] = static::MFAPOLLPENDING;
                } elseif ($rejected) {
                    $results[static::MFASTATUS] = static::MFAFAILED;
                }

                return $this->prepareSuccessModel($results);
            }
        }
        // If we hit here then we have failed and should return the error if we have one.
        return new JsonModel(
            array(
                'isValid'         => false,
                static::CODE      => Response::STATUS_CODE_401,
                static::ERROR     => $message,
                static::MFASTATUS => static::MFAFAILED
            )
        );
    }

    /**
     * Take an array of methods element sort them into separate array within an array
     *
     * @param $methods    Single dimension array
     * @return array      Multi dimension array
     */
    protected function listMethodSort($methods)
    {
        $results = array();
        $last    = 0; // Keep track of which method we last saw

        // loop though each of the data sent back and create an array to send in json
        foreach ($methods as $key => $method) {
            preg_match("/(.*)([0-9]+)$/U", $key, $spilt);
            if (isset($spilt[1]) && isset($spilt[2])) {
                $field   = $spilt[1]; // This is the name of the field
                $current = $spilt[2]; // This is the current method number we are at

                // If we are still on the same method then add the next field to it.
                if ($current === $last) {
                    $results[$current][$field] = $method;
                } else {
                    // else we set current and create the next method in the array.
                    $last = $current;

                    $results[$current][$field] = $method;
                }
            }
        }
        return $results;
    }
}
