<?php
/**
 * Manager to help with getting values from a configuration array.
 *
 * @copyright   2013-2018 Perforce Software. All rights reserved.
 * @license     Please see LICENSE.txt in top-level readme folder of this distribution.
 * @version     2018.1/1742263
 */

namespace Application\Config;

use P4\Time\Time;
use Users\Settings\ReviewPreferences;
use Users\Settings\TimePreferences;
use Reviews\Model\Review;

class ConfigManager
{
    const DIFF_MAX_DIFFS                          = 'diffs.max_diffs';
    const FILES_DOWNLOAD_TIMEOUT                  = 'files.download_timeout';
    const FILES_MAX_SIZE                          = 'files.max_size';
    const REVIEWS_EXPAND_ALL                      = 'reviews.expand_all_file_limit';
    const REVIEWS_DISABLE_TESTS_ON_APPROVE_COMMIT = 'reviews.disable_tests_on_approve_commit';
    const REVIEWS_DISABLE_APPROVE_WHEN_TASKS_OPEN = 'reviews.disable_approve_when_tasks_open';
    const REVIEWS_EXPAND_GROUP_REVIEWERS          = 'reviews.expand_group_reviewers';
    const REVIEWS_FILTERS_RESULT_SORTING          = 'reviews.filters.result_sorting';
    const REVIEWS_FILTERS_DATE_FIELD              = 'reviews.filters.date_field';
    const UPGRADE_BATCH_SIZE                      = 'upgrade.batch_size';
    const UPGRADE_STATUS_REFRESH_INTERVAL         = 'upgrade.status_refresh_interval';
    const AVATARS_HTTP                            = 'avatars.http_url';
    const AVATARS_HTTPS                           = 'avatars.https_url';
    const SECURITY_REQUIRE_LOGIN                  = 'security.require_login';
    const WORKER_CHANGE_SAVE_DELAY                = 'queue.worker_change_save_delay';
    const TRANSLATOR_NON_UTF8_ENCODINGS           = 'translator.non_utf8_encodings';
    const PROJECTS_SIDEBAR_FOLLOWERS_DISABLED     = 'projects.sidebar_followers_disabled';
    const DASHBOARD_MAX_ACTIONS                   = 'users.maximum_dashboard_actions';
    const PROJECTS_README_MODE                    = 'projects.readme_mode';
    const PROJECTS_MAX_README_SIZE                = 'projects.max_readme_size';
    const PROJECTS_MAINLINES                      = 'projects.mainlines';
    const COMMENT_SHOW_ID                         = 'comments.show_id';
    const COMMENT_THREADING_MAX_DEPTH             = 'comments.threading.max_depth';
    const COMMENT_NOTIFICATION_DELAY_TIME         = 'comments.notification_delay_time';
    const JIRA_LINK_TO_JOBS                       = 'jira.link_to_jobs';
    const JIRA_DELAY_JOB_LINKS                    = 'jira.delay_job_links';
    const REVIEWS_PROCESS_SHELF_DELETE_WHEN       = 'reviews.process_shelf_delete_when';
    const MORE_CONTEXT_LINES                      = 'more_context_lines';
    const REVIEWS_MORE_CONTEXT_LINES              = 'reviews.more_context_lines'; // 5.3 cannot concat const
    const MAX_BOTTOM_CONTEXT_LINES                = 'max_bottom_context_lines';
    const REVIEWS_MAX_BOTTOM_CONTEXT_LINES        = 'reviews.max_bottom_context_lines'; // 5.3 cannot concat const

    // User setting.
    const USER_SETTINGS = 'users.settings';
    // User preferences settings.
    const USER_SETTINGS_REVIEW_PREF_SHOW_COMMENTS     = "users.settings.review_preferences.show_comments_in_files";
    const USER_SETTINGS_REVIEW_PREF_SIDE_BY_SIDE      = "users.settings.review_preferences.view_diffs_side_by_side";
    const USER_SETTINGS_REVIEW_PREF_IGNORE_WHITESPACE = "users.settings.review_preferences.ignore_whitespace";
    const USER_SETTINGS_TIME_DISPLAY                  = 'users.settings.time.display';
    // Due to php5.3 not liking constants being put together in a new constant.
    const USER_SETTINGS_REVIEW_PREF_SHOW_SPACE = "users.settings.review_preferences.show_space_and_new_line_characters";

    const SETTINGS = 'settings';

    const STRING           = 'string';
    const INT              = 'int';
    const INT_ALLOW_NULL   = 'intAllowNull';
    const BOOLEAN          = 'boolean';
    const ARRAY_OF_STRINGS = 'arrayOfStrings';

    // Case sensitivity levels
    const CASE_SENSITIVITY = 'case_sensitivity';
    const CASE_INSENSITIVE = 1;
    const CASE_SENSITIVE   = 0;

    // Metadata to define types etc (structure should agree with the configuration array)
    private static $configMetaData = array(
        'comments' => array(
            'threading' => array(
                'max_depth' => array('type' => self::INT)
            ),
            'show_id'                 => array('type' => self::BOOLEAN),
            'notification_delay_time' => array('type' => self::INT)
        ),
        'diffs'   => array(
            'max_diffs'             => array('type' => self::INT),
        ),
        'files'   => array(
            'max_size'              => array('type' => self::INT),
            'download_timeout'      => array('type' => self::INT)
        ),
        'reviews' => array(
            'expand_all_file_limit' => array('type' => self::INT),
            'filters'               => array(
                'result_sorting'    => array('type' => self::BOOLEAN),
                'date_field'        => array('type' => self::STRING, 'valid_values' => array('created', 'updated'))
            ),
            'expand_group_reviewers'          => array('type' => self::BOOLEAN),
            'disable_tests_on_approve_commit' => array('type' => self::BOOLEAN),
            'disable_approve_when_tasks_open' => array('type' => self::BOOLEAN),
            'process_shelf_delete_when'       => array(
                'type'                 => self::ARRAY_OF_STRINGS,
                'valid_values'         => array(
                    Review::STATE_APPROVED,
                    Review::STATE_APPROVED_COMMIT,
                    Review::STATE_ARCHIVED,
                    Review::STATE_NEEDS_REVIEW,
                    Review::STATE_NEEDS_REVISION,
                    Review::STATE_REJECTED
                ),
                self::CASE_SENSITIVITY => self::CASE_SENSITIVE
            ),
            self::MORE_CONTEXT_LINES => array('type' => self::INT),
            self::MAX_BOTTOM_CONTEXT_LINES => array('type' => self::INT)
        ),
        'projects' => array(
            'sidebar_followers_disabled' => array('type' => self::BOOLEAN),
            'readme_mode' =>
                array('type'  => self::STRING, 'valid_values' => array('disabled', 'restricted', 'unrestricted')),
            'mainlines'       => array('type' => self::ARRAY_OF_STRINGS),
            'max_readme_size' => array('type' => self::INT_ALLOW_NULL)
        ),
        'upgrade' => array(
            'status_refresh_interval' => array('type' => self::INT),
            'batch_size'              => array('type' => self::INT)
        ),
        'avatars' => array(
            'http_url'  => array(
                'type' => self::STRING,
                self::CASE_SENSITIVITY => self::CASE_SENSITIVE
            ),
            'https_url' => array(
                'type' => self::STRING,
                self::CASE_SENSITIVITY => self::CASE_SENSITIVE
            )
        ),
        'security' => array(
            'require_login' => array('type' => self::BOOLEAN)
        ),
        'queue' => array(
            'worker_change_save_delay' => array('type' => self::INT)
        ),
        'translator' => array(
            'non_utf8_encodings' => array('type' => self::ARRAY_OF_STRINGS)
        ),
        'users' => array(
            'maximum_dashboard_actions' => array('type' => self::INT),
            self::SETTINGS => array(
                ReviewPreferences::REVIEW_PREFERENCES => array(
                    ReviewPreferences::SHOW_COMMENTS_IN_FILES  => array('type' => self::BOOLEAN),
                    ReviewPreferences::VIEW_DIFFS_SIDE_BY_SIDE => array('type' => self::BOOLEAN),
                    ReviewPreferences::SHOW_SPACE_AND_NEW_LINE => array('type' => self::BOOLEAN),
                    ReviewPreferences::IGNORE_WHITESPACE       => array('type' => self::BOOLEAN)
                ),
                TimePreferences::TIME_PREFERENCES => array (
                    TimePreferences::DISPLAY  => array(
                        'type' => self::STRING,
                        'valid_values' => array(TimePreferences::TIMEAGO, TimePreferences::TIMESTAMP)
                    )
                )
            ),
        ),
        'jira' => array(
            'link_to_jobs'    => array('type' => self::BOOLEAN),
            'delay_job_links' => array('type' => self::INT)
        )
    );

    /**
     * Gets a value from the config and checks if it is valid. Simple case at the moment - we may
     * want to enhance to add ranges etc.
     * @param $config array config data
     * @param $path string array path (for nested values use dot notation, for example 'reviews.expand_all_file_limit'.
     * @param null $default default value that will be returned without validation if any problems occur
     * @param null $metaData metadata override
     * @return array|bool|int|mixed|null|string
     * @throws ConfigException
     * @throws \Exception
     */
    public static function getValue($config, $path, $default = null, $metaData = null)
    {
        // This will throw an exception if the value is not set
        $configValue     = null;
        $metaDataElement = null;
        try {
            $configValue = self::getValueFromConfig($config, $path);
            try {
                $metaDataElement = self::getValueFromConfig($metaData ? $metaData : self::$configMetaData, $path);
            } catch (ConfigException $e) {
                // Something is wrong with our metadata, perhaps a path is asked for that we haven't yet catered for.
                // Don't fail because of this.
                return $configValue;
            }
            // This will throw an exception if validation fails
            $configValue = self::validate($configValue, $metaDataElement, $path);
        } catch (ConfigException $ce) {
            if ($default === null || $metaDataElement === null) {
                throw $ce;
            } else {
                $configValue = self::validateDefault($default, $metaDataElement);
            }
        }
        return $configValue;
    }

    /**
     * Make sure that if a default is provided it fits the metadata definition.
     * @param $value
     * @param $metaDataValue
     * @return null
     * @throws ConfigException
     */
    private static function validateDefault($value, $metaDataValue)
    {
        $type    = $metaDataValue['type'];
        $default = null;
        switch ($type) {
            case self::STRING:
                if (isset($metaDataValue['valid_values']) &&
                    !array_uintersect(array($value), $metaDataValue['valid_values'], 'strcasecmp')) {
                    throw new ConfigException(
                        "Value '" . $value . "' provided as a default must be one of '"
                        . implode(', ', $metaDataValue['valid_values'])
                    );
                }
                $default = strtolower($value);
                break;
            default:
                $default = $value;
                break;
        }
        return $default;
    }

    /**
     * Validates the value.
     * @param $value mixed value to validate
     * @param $metaDataValue string meta data
     * @param $path string the path
     * @return array|bool|int|null|string
     * @throws ConfigException
     */
    private static function validate($value, $metaDataValue, $path)
    {
        // Validate all parameters here - currently just type is specified
        $type           = isset($metaDataValue['type']) ? $metaDataValue['type'] : null;
        $convertedValue = null;
        $nullPermitted  = false;
        switch ($type) {
            case self::INT:
                if (ctype_digit(strval($value))) {
                    $convertedValue = intval($value);
                }
                break;
            case self::BOOLEAN:
                if ($value === true || strcasecmp($value, 'true') == 0) {
                    $convertedValue = true;
                } elseif ($value === false || strcasecmp($value, 'false') == 0) {
                    $convertedValue = false;
                }
                break;
            case self::STRING:
                if (is_string($value)) {
                    if (isset($metaDataValue[self::CASE_SENSITIVITY])
                        && $metaDataValue[self::CASE_SENSITIVITY] === self::CASE_SENSITIVE) {
                        $convertedValue = $value;
                    } else {
                        $convertedValue = strtolower($value);
                    }
                    if (isset($metaDataValue['valid_values']) &&
                        !array_uintersect(array($value), $metaDataValue['valid_values'], 'strcasecmp')) {
                        // valid values are specified and the value does not match them
                        $convertedValue = null;
                    }
                }
                break;
            case self::ARRAY_OF_STRINGS:
                // We support setting a single string value that we will convert to an array
                $value = is_string($value) ? (array) $value : $value;
                if (is_array($value)) {
                    $convertedValue = array();
                    foreach ($value as $arrayValue) {
                        if (is_string($arrayValue)) {
                            // Check if the case sensitivity enabled.
                            if (isset($metaDataValue[self::CASE_SENSITIVITY])
                                && $metaDataValue[self::CASE_SENSITIVITY] === self::CASE_SENSITIVE) {
                                $convertedValue[] = $arrayValue;
                                // Now also compare the array in case sensitive mode.
                                if (isset($metaDataValue['valid_values']) &&
                                    !array_intersect(array($arrayValue), $metaDataValue['valid_values'])) {
                                    // valid values are specified and the value does not match them
                                    $convertedValue = null;
                                    break;
                                }
                            } else {
                                $convertedValue[] = strtolower($arrayValue);
                                if (isset($metaDataValue['valid_values']) &&
                                    !array_uintersect(
                                        array($arrayValue),
                                        $metaDataValue['valid_values'],
                                        'strcasecmp'
                                    )
                                ) {
                                    // valid values are specified and the value does not match them
                                    $convertedValue = null;
                                    break;
                                }
                            }
                        } else {
                            $convertedValue = null;
                            break;
                        }
                    }
                }
                break;
            case self::INT_ALLOW_NULL:
                if ($value === null) {
                    $nullPermitted = true;
                } else {
                    if (ctype_digit(strval($value))) {
                        $convertedValue = intval($value);
                    }
                }
                break;
            case null:
                $convertedValue = $value;
                break;
        }
        if ($convertedValue === null && $nullPermitted !== true) {
            throw new ConfigException(
                "Value '" . (is_array($value) ? var_export($value, true) : $value) .
                "' at path '" . $path . "' is invalid"
            );
        }
        return $convertedValue;
    }

    /**
     * Iterates through the configuration to find a value.
     * @param $config array config
     * @param $path string array path (for nested values use dot notation, for example 'reviews.expand_all_file_limit'.
     * @return mixed
     * @throws \Exception if the path does not exist or the configuration being searched is not an array.
     */
    private static function getValueFromConfig($config, $path)
    {
        $ref  = &$config;
        $keys = explode('.', $path);
        foreach ($keys as $idx => $key) {
            if (!is_array($ref)) {
                throw new ConfigException('Configuration is not an array');
            }
            if (!array_key_exists($key, $ref)) {
                throw new ConfigException("Path '" . $path . "' does not exist");
            }
            $ref = &$ref[$key];
        }
        return $ref;
    }
}
