<?php
/**
 * Perforce Swarm
 *
 * @copyright   2013-2018 Perforce Software. All rights reserved.
 * @license     Please see LICENSE.txt in top-level readme folder of this distribution.
 * @version     2018.1/1742263
 */

namespace Jobs;

use Activity\Model\Activity;
use Application\Filter\Linkify;
use Mail\Module as Mail;
use P4\Log\Logger;
use Projects\Model\Project;
use Notifications\Settings;
use Users\Model\User;
use Zend\Mvc\MvcEvent;

class Module
{
    /**
     * Connect to queue event manager to handle jobs.
     *
     * @param   MvcEvent    $event  the bootstrap event
     * @return  void
     */
    public function onBootstrap(MvcEvent $event)
    {
        $application = $event->getApplication();
        $services    = $application->getServiceManager();
        $events      = $services->get('queue')->getEventManager();
        $jobModule   = $this;
        // fetch job object for job events
        $events->attach(
            'task.job',
            function ($event) use ($services, $jobModule) {
                $p4Admin = $services->get('p4_admin');
                $id      = $event->getParam('id');

                try {
                    $job = \P4\Spec\Job::fetchById($id, $p4Admin);
                    $event->setParam('job', $job);

                    // determine event author
                    // by default there is no modified-by field, but if we
                    // can find one in the jobspec, we will use it here.
                    $user = $job->hasModifiedByField()
                        ? $job->get($job->getModifiedByField())
                        : $job->getUser();

                    // determine action the user took
                    $action = 'modified';
                    if ($job->hasCreatedDateField() && $job->hasModifiedDateField()) {
                        $created  = $job->get($job->getCreatedDateField());
                        $modified = $job->get($job->getModifiedDateField());
                        $action   = $created === $modified ? 'created' : 'modified';
                    }

                    // prepare data model for activity streams
                    $activity = new Activity;
                    $activity->set(
                        array(
                            'type'          => 'job',
                            'link'          => array('job', array('job' => $job->getId())),
                            'user'          => $user,
                            'action'        => $action,
                            'target'        => $job->getId(),
                            'description'   => $job->getDescription(),
                            'topic'         => 'jobs/' . $job->getId(),
                            'time'          => $event->getParam('time'),
                            'projects'      => Project::getAffectedByJob($job, $p4Admin)
                        )
                    );
                    // ensure any @mention'ed users are included
                    $mentions = User::filter(Linkify::getCallouts($job->getDescription()), $p4Admin);
                    $activity->addFollowers($mentions);

                    $event->setParam('activity', $activity);

                    // Now check if we need to notify anyone for job creation or edits.
                    $jobModule->checkJobNotification($job, $user, $event, $services);
                } catch (\Exception $e) {
                    $services->get('logger')->err($e);
                }
            },
            100
        );
    }

    /**
     * To incorporate the p4review2 feature that notify people if a job is created or
     * edited it should send them an email.
     *
     * @param $job     The job that triggered the event.
     * @param $user    The user that did the creation or edit of job.
     * @param $event   The full event module.
     * @param $services The services object.
     */
    public function checkJobNotification($job, $user, $event, $services)
    {
        $config  = $services->get('config');
        $p4Admin = $services->get('p4_admin');
        // Get the notification settings.
        $notifications = isset($config[Settings::NOTIFICATIONS])
            ? $config[Settings::NOTIFICATIONS]
            : array();
        // Check if the opt_in_job_path has been set otherwise set it to null
        $notifications += array(
            Settings::OPT_IN_JOB_PATH    => null,
        );

        // Now check that review path to see who wants emails for the job changes or creation.
        $reviewPath = $notifications[Settings::OPT_IN_JOB_PATH];
        if ($reviewPath && is_string($reviewPath)) {
            // Find who wants an email about jobs.
            $users   = $p4Admin->run('reviews', array($reviewPath))->getData();
            $toUsers = array();
            foreach ($users as $userDetails) {
                $toUsers[] = $userDetails['user'];
            }
            // ensure we don't have any duplicate users.
            array_unique($toUsers);
            // Debug out who we are attempting to send the job email to.
            $logger = $services->get('logger');
            $logger->debug("Job/Mail: We are preparing a job email for " . var_export($toUsers, true));
            // configure a message for mail module to deliver
            $mailParams = array(
                'subject'       => 'Job '. $job->getId() . ' for review',
                'author'        => $user,
                'toUsers'       => $toUsers,
                'fromUser'      => $user,
                'job'           => $job,
                'messageId'     =>
                    '<topic-job/' . $job->getId()
                    . '@' . Mail::getInstanceName($config['mail']) . '>',
                'htmlTemplate'  => __DIR__ . '/view/mail/job-html.phtml',
                'textTemplate'  => __DIR__ . '/view/mail/job-text.phtml',
            );
            // Set the mail param on the event
            $event->setParam('mail', $mailParams);
        }
    }

    public function getConfig()
    {
        return include __DIR__ . '/config/module.config.php';
    }

    public function getAutoloaderConfig()
    {
        return array(
            'Zend\Loader\StandardAutoloader' => array(
                'namespaces' => array(
                    __NAMESPACE__ => __DIR__ . '/src/' . __NAMESPACE__,
                ),
            ),
        );
    }
}
