<?php
/**
 * Perforce Swarm
 *
 * @copyright   2013-2018 Perforce Software. All rights reserved.
 * @license     Please see LICENSE.txt in top-level readme folder of this distribution.
 * @version     2018.1/1742263
 */

namespace Reviews;

use Application\Config\ConfigManager;
use P4\Exception;
use P4\Spec\Change;
use Record\Exception\NotFoundException as RecordNotFoundException;
use Reviews\Listener\Review as ReviewListener;
use Reviews\Listener\ShelveCommit as ShelveCommitListener;
use Reviews\Filter\GitInfo;
use Reviews\Model\GitReview;
use Reviews\Model\Review;
use Zend\Mvc\MvcEvent;

class Module
{
    /**
     * Connect to queue event manager to handle review tasks.
     *
     * @param   MvcEvent    $event  the bootstrap event
     * @return  void
     */
    public function onBootstrap(MvcEvent $event)
    {
        $application  = $event->getApplication();
        $services     = $application->getServiceManager();
        $events       = $services->get('queue')->getEventManager();
        $reviewModule = $this;

        // attach listeners to:
        // - create/update review when change is shelved or committed
        // - process review when its created or updated
        $events->attach(new ShelveCommitListener($services));
        $events->attach(new ReviewListener($services));

        // Deal with git-fusion initiated reviews before traditional p4 reviews.
        //
        // If the shelf is already a git initiated review we'll update it.
        //
        // If the shelf has git-fusion keywords indicating this is a new review
        // translate the existing shelf into a new git-review and process it.
        $events->attach(
            'task.shelve',
            function ($event) use ($services) {
                $p4Admin = $services->get('p4_admin');
                $queue   = $services->get('queue');
                $config  = $services->get('config');
                $change  = $event->getParam('change');

                // if we didn't get a pending change to work with, bail
                if (!$change instanceof Change || !$change->isPending()) {
                    return;
                }

                // if the change is by a user that is ignored for the purpose of reviews, bail
                $ignored = isset($config['reviews']['ignored_users']) ? $config['reviews']['ignored_users'] : null;
                if ($p4Admin->stringMatches($change->getUser(), (array) $ignored)) {
                    return;
                }

                // if this change doesn't have a valid looking git-fusion style review-id
                // there's no need to further examine it here, return
                $gitInfo = new GitInfo($change->getDescription());
                if ($gitInfo->get('review-id') != $change->getId()) {
                    return;
                }

                try {
                    // using the change id, verify if a git review already exists
                    // note the review id and change id are the same for git-fusion reviews
                    $review = Review::fetch($change->getId(), $p4Admin);

                    // if we get a review but its the wrong type, we can't do anything with it
                    // this really shouldn't happen but good to confirm all is well
                    if ($review->getType() != 'git') {
                        return;
                    }
                } catch (RecordNotFoundException $e) {
                    // couldn't fetch an existing review, create one!
                    $review = GitReview::createFromChange($change, $p4Admin);
                    $review->save();

                    // ensure we pass along to the review event that this is an add
                    $isAdd = true;
                }

                // put the fetched/created review on the existing event.
                // the presence of a review on the event will cause the traditional
                // shelf-commit handler to skip processing this change.
                $event->setParam('review', $review);

                // push the new review into queue for further processing, always replace for git reviews.
                $queue->addTask(
                    'review',
                    $review->getId(),
                    array(
                        'user'                  => $change->getUser(),
                        'updateFromChange'      => $change->getId(),
                        Review::ADD_CHANGE_MODE => Review::REPLACE_MODE,
                        'isAdd'                 => isset($isAdd) && $isAdd
                    )
                );
            },
            100
        );

        // This will only fire if we have a task.shelvedelete event that has action.
        $events->attach(
            'task.shelvedel',
            function ($event) use ($services, $reviewModule) {
                $p4Admin = $services->get('p4_admin');
                $queue   = $services->get('queue');
                $config  = $services->get('config');
                $logger  = $services->get('logger');

                // We need to know who is running Swarm to exclude that user from
                // triggering this work
                $userRunningSwarm = isset($config['p4'][Review::USER])
                                    ? $config['p4'][Review::USER]
                                    : null;

                $dataValues = $event->getParam('data');
                $user       = isset($dataValues[Review::USER]) && !empty($dataValues[Review::USER])
                              ? $dataValues[Review::USER]
                              : null;
                // Swarm cleans its own changelist up a lot with the shelve -d so we want to
                // exclude the user in the config from firing this action. Exit as early as possible
                if ($user && $user === $userRunningSwarm) {
                    return;
                }

                $id     = $event->getParam(Review::FIELD_ID);
                $client = isset($dataValues[Review::CLIENT]) && !empty($dataValues[Review::CLIENT])
                    ? $dataValues[Review::CLIENT] : null;
                $cwd    = isset($dataValues[Review::CWD])    && !empty($dataValues[Review::CWD])
                    ? $dataValues[Review::CWD]    : null;
                $files  = isset($dataValues[Review::FILES])  && !empty($dataValues[Review::FILES])
                    ? $dataValues[Review::FILES]  : null;

                if (is_null($client) || is_null($cwd) || is_null($files)|| is_null($user)) {
                    $message = sprintf(
                        "%s Unable to process id[%s] for values client[%s], working directory[%s], files[%s], " .
                        "user[%s]. This is not a problem if the delete was triggered by deleting the entire shelf.",
                        Review::SHELVEDEL,
                        $id,
                        var_export($client, true),
                        var_export($cwd, true),
                        var_export($files, true),
                        var_export($user, true)
                    );
                    // We can't process the delete if we don't have these required field.
                    $logger->warn($message);
                    return;
                }

                // Try and fetch the review for the given Change id.
                $review = Review::fetchAll(
                    array(
                        Review::FETCH_BY_CHANGE => $id
                    ),
                    $p4Admin
                )->first();

                if ($review === false) {
                    // Calling 'first' on the iterator above will result in false being returned when there are no
                    // items. This is a valid scenario if the change has no review
                    $logger->trace(sprintf('%s No review found to process for change [%s]', Review::SHELVEDEL, $id));
                } else {
                    // if we get a review but its the git type, ignore it as we don't want
                    // handle them.
                    if ($review->getType() === 'git') {
                        $logger->trace(Review::SHELVEDEL . 'We will not process shelvedel for a git review.');
                        return;
                    }
                    // Get the states that we can process.
                    $defaultExcludedStates = array(Review::STATE_ARCHIVED, Review::STATE_APPROVED_COMMIT);
                    $configStates          = ConfigManager::getValue(
                        $config,
                        ConfigManager::REVIEWS_PROCESS_SHELF_DELETE_WHEN
                    );


                    // If the current review state isn't a state we allowed to process end now.
                    // By default we don't allow review States archived or approved:commit
                    if (!in_array($review->getState(), $configStates)
                        || in_array($review->getState(), $defaultExcludedStates)) {
                        $logger->trace(Review::SHELVEDEL . 'We have not found an allowed state to be processed.');
                        return;
                    }
                    // Now we need to go off and check which files this event is trying to delete.
                    $depotFiles = $reviewModule->getFileDepotPath($services, $client, $cwd, $files);

                    // Only raise a review task if we have some depot files to process. Atomic
                    // delete of a shelf passing no files should not update the review
                    if ($depotFiles && sizeof($depotFiles) > 0) {
                        // put the fetched review on the existing event.
                        $event->setParam('review', $review);

                        // push the task into queue for further processing.
                        $queue->addTask(
                            'review',
                            $review->getId(),
                            array(
                                Review::USER => $user,
                                Review::DELFROMCHANGE => $review->getId(),
                                Review::FILES => $depotFiles,
                                Review::CLIENT => $client
                            )
                        );
                    }
                }
            }
        );
    }

    /**
     * This is given a list of files and relative path and we need to
     * build and work out the depot path.
     *
     * Data from trigger will be like this:
     * suzie-photon-test,/home/spenn/tmp/suzie-photon/depot,-c,145697,-d,b.txt,c.txt
     *
     * Which is broken down into:
     * Client
     * Relative path
     * -c
     * changelist
     * -d
     * files ... (the rest should be all the individual files.)
     *
     * @param $services This is the service locator.
     * @param $client   This is the client the files came from.
     * @param $path     This is the relative path for all files.
     * @param $files    This is the list of files.
     * @return mixed
     */
    public function getFileDepotPath($services, $client, $path, $files)
    {
        if (!is_array($files)) {
            $files = array($files);
        }
        $p4Admin = $services->get('p4_admin');
        $logger  = $services->get('logger');

        // Keep the host to be reset back later.
        $oldHost = $p4Admin->getHost();
        try {
            // Try and fetch the client the user is using.
            $clientData = $p4Admin->run('client', array('-o', $client));
            $clientData = $clientData->getData(0);
            $host       = isset($clientData['Host']) ? $clientData['Host'] : $oldHost;
            // Set the p4admin connection to be able to use the users client.
            $p4Admin->setHost($host);
            $p4Admin->setClient($client);
        } catch (Exception $clientError) {
            $logger->err(
                Review::SHELVEDEL. "There was an setting up the client and host. Error: " . $clientError->getMessage()
            );
        }

        $depotFiles = array();
        foreach ($files as $file) {
            // If the file we have been provide is a depot path just accepted it.
            if (substr($file, 0, 2) === "//") {
                $depotFiles[] = $file;
            } else {
                //The relativePath will be given to us by the trigger.
                $relativeFile = $path . '/' . $file;
                // Know we have the users client and host set we can do a p4 where files
                // which will return us the depot path of them files.
                try {
                    // If the 'where' fails carry on with other files but just log and move on
                    // See SW-5250 for issue description
                    $depotLoc     = $p4Admin->run('where', $relativeFile);
                    $newData      = $depotLoc->getData(0);
                    $depotFiles[] = $newData['depotFile'];
                } catch (Exception $error) {
                    $logger->trace(Review::SHELVEDEL . "Failed to get where output for file: " . $error->getMessage());
                }
            }
        }

        // Set the old Host name back incase we have locked clients.
        $p4Admin->setHost($oldHost);

        return $depotFiles;
    }

    public function getConfig()
    {
        return include __DIR__ . '/config/module.config.php';
    }

    public function getAutoloaderConfig()
    {
        return array(
            'Zend\Loader\StandardAutoloader' => array(
                'namespaces' => array(
                    __NAMESPACE__ => __DIR__ . '/src/' . __NAMESPACE__,
                ),
            ),
        );
    }
}
